from typing import Any, Dict, List, Optional

from .db import ConnectionPoolDB


class MultipleRecordsFoundError(Exception):
    """
    Thrown if multiple records are found in the database
    and there should be one.
    """

    pass


class DataAccessMixin:
    def __init__(self, db: ConnectionPoolDB):
        self.db = db

    def _load_item_from_database(
        self,
        table: str,
        columns: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        where_operator: str = "=",
    ) -> Any:
        """
        Loads a single item from the database.

        :param table: Name of the table.
        :param columns: List of columns to select (defaults to all columns).
        :param where: WHERE conditions as a dictionary.
        :return: item as a dictionary.
        """
        try:
            res = self.db.select(
                table=table, columns=columns, where=where, where_operator=where_operator
            )
        except Exception:
            raise
        count = len(res)
        if count == 0:
            return None
        if count != 1:
            raise MultipleRecordsFoundError(f"Found {count} records in {table}")
        return res[0]

    def _load_user(
        self,
        where: Dict[str, Any],
        where_operator: str = "=",
        columns: List[str] = [
            "user_id",
            "username",
            "pwhash",
            "groups",
            "hashfn",
            "display_name",
            "certhash",
            "last_connect",
            "last_disconnect",
            "release",
            "version",
        ],
    ) -> Any:
        """
        Loads a user from the database.

        Database table:
        `user_id` int(11) NOT NULL,
        `username` varchar(254) NOT NULL, UNIQUE
        `pwhash` varchar(90) NOT NULL,
        `groups` longtext DEFAULT NULL,
        `hashfn` varchar(20) NOT NULL,
        `display_name` varchar(254) NOT NULL, UNIQUE
        `certhash` varchar(254) DEFAULT NULL,
        `last_connect` datetime(6) DEFAULT NULL,
        `last_disconnect` datetime(6) DEFAULT NULL,
        `release` longtext DEFAULT NULL,
        `version` int(11) DEFAULT NULL,

        :param name: Name of the user
        """
        return self._load_item_from_database(
            table="mumble_mumbleuser", columns=columns, where=where, where_operator=where_operator
        )

    def _load_temp_user(
        self,
        where: Dict[str, Any],
        where_operator: str = "=",
        columns: List[str] = [
            "id",
            "name",
            "username",
            "password",
            "expires",
            "templink_id",
            "character_id",
        ],
    ) -> Any:
        """
        Loads a temp user from the database.

        Database table:
        `id` int(11) NOT NULL AUTO_INCREMENT,
        `name` varchar(200) NOT NULL,
        `username` varchar(20) NOT NULL, UNIQUE
        `password` varchar(20) NOT NULL,
        `expires` int(11) NOT NULL,
        `templink_id` int(11) DEFAULT NULL,
        `character_id` int(11) NOT NULL,

        :param where: WHERE conditions as a dictionary.
        :param columns: List of columns to select
        (defaults: id, name, username, password, expires, templink_id, character_id).
        """
        return self._load_item_from_database(
            table="mumbletemps_tempuser",
            columns=columns,
            where=where,
            where_operator=where_operator,
        )



from logging import debug, error

from passlib.hash import bcrypt, bcrypt_sha256


def allianceauth_check_hash(password, hash, hash_type):
    """
    Python implementation of the AllianceAuth MumbleUser hash function

    See https://gitlab.com/allianceauth/allianceauth/-/blob/master/allianceauth/services/modules/mumble/models.py
    :param password: Password to be verified
    :param hash: Hash for the password to be checked against
    :param hash_type: Hashing function originally used to generate the hash
    """
    debug(f"Checking password with hash type: {hash_type}")
    if not password:
        error("Empty password")
        return False
    if not hash:
        error("Empty hash")
        return False
    try:
        if hash_type == "bcrypt-sha256":
            return bcrypt_sha256.verify(password, hash)
        else:
            return bcrypt.verify(password, hash)
    except Exception as e:
        error(f"Failed to verify password hash: {e}")
        return False



import inspect
import time
from html import unescape
from logging import debug, error, info, warning
from typing import Any, Dict, List, Optional, Tuple

from eveo7_mumbleserver_ice import ServerUpdatingAuthenticator, UserInfo
from prometheus_client import Enum, Summary

from .db import ConnectionPoolDB
from .decorators import check_secret, fortify_ice
from .mixin_avatarcache import AvatarCacheMixin
from .mixin_dataaccess import DataAccessMixin
from .mixin_idconversion import IDConversionMixin
from .password import allianceauth_check_hash
from .state import (
    AuthenticationState,
    GetInfoState,
    IdToNameState,
    IdToTextureState,
    NameToIdState,
    RegisterUserState,
    SetInfoState,
    SetTextureState,
    UnregisterUserState,
)

USER_AUTHENTICATION_STATE = Enum(
    "user_authentication_state",
    "Authentication state of users",
    labelnames=["username"],
    states=["success", "failure", "fallthrough", "temporarily"],
)

USER_AUTHENTICATION_LATENCY = Summary(
    "user_authentication_latency",
    "Time taken to authenticate users",
)


class Authenticator(
    ServerUpdatingAuthenticator, IDConversionMixin, DataAccessMixin, AvatarCacheMixin
):
    def __init__(self, cfg: Dict[str, Any], db: ConnectionPoolDB):
        super().__init__()
        DataAccessMixin.__init__(self, db=db)
        IDConversionMixin.__init__(self, base_offset=cfg["user"]["id_offset"])
        AvatarCacheMixin.__init__(
            self,
            cache_dir=cfg["avatar"]["cache_dir"],
            cache_ttl=cfg["avatar"]["cache_ttl"],
            cache_max_size=cfg["avatar"]["cache_max_size"],
            cache_max_age=cfg["avatar"]["cache_max_age"],
            url_template=cfg["avatar"]["url"],
        )
        self.db = db
        self.cfg = cfg
        self._decorate_methods()

    def _decorate_methods(self) -> None:
        """
        Decorates class methods with security checks and fallback return values.
        """
        secret = self.cfg["ice_server"]["secret"]
        fallback_results = {
            "authenticate": (
                (AuthenticationState.FAILURE.value, None, None)
                if self.cfg["user"]["reject_on_error"]
                else (AuthenticationState.FALLTHROUGH.value, None, None)
            ),
            "getInfo": (GetInfoState.FALLTHROUGH.value, None),
            "nameToId": NameToIdState.UNKNOWN.value,
            "registerUser": RegisterUserState.FALLTHROUGH.value,
            "idToName": IdToNameState.UNKNOWN.value,
            "idToTexture": IdToTextureState.UNKNOWN.value,
            "getRegisteredUsers": {},
        }
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not name.startswith("_"):
                fallback_result = fallback_results.get(name)
                # debug(f"Decorating {name} in Authenticator with fallback {fallback_result}")
                if fallback_result:
                    decorated_method = fortify_ice(retval=fallback_result)(
                        check_secret(secret=secret)(method)
                    )
                else:
                    decorated_method = check_secret(secret=secret)(method)
                setattr(self, name, decorated_method)

    @staticmethod
    def _user_authentication_state(username: str, state: str) -> None:
        """
        Sets the prometheus metric for the user authentication state

        :param username: The username of the user
        :param state: The authentication state
        """
        USER_AUTHENTICATION_STATE.labels(username=username).state(state)

    def _authenticate_fallthrough(self, username: str) -> Tuple[int, None, None]:
        """
        Helper method to set authentication state to fallthrough

        :param username: The username of the user
        :return: The authentication state
        """
        self._user_authentication_state(username, "fallthrough")
        return (AuthenticationState.FALLTHROUGH.value, None, None)

    def _authenticate_failure(self, username: str) -> Tuple[int, None, None]:
        """
        Helper method to set authentication state to failure

        :param username: The username of the user
        :return: The authentication state
        """
        self._user_authentication_state(username, "failure")
        return (AuthenticationState.FAILURE.value, None, None)

    def _authenticate_success(
        self, username: str, mumble_user_id: int, display_name: str, groups: List[str]
    ) -> Tuple[int, str, List[str]]:
        """
        Helper method to set authentication state to success

        :param username: The username of the user
        :param mumble_user_id: The mumble user id
        :param display_name: The display name
        :param groups: The groups list
        :return: The authentication state tuple
        """
        self._user_authentication_state(username, "success")
        return (
            mumble_user_id,
            unescape(display_name),
            groups,
        )

    @USER_AUTHENTICATION_LATENCY.time()
    def authenticate(
        self, name, pw, certlist, certhash, strong, current=None
    ) -> Tuple[int, Optional[str], Optional[List[str]]]:
        """
        Called to authenticate a user. If you do not know the username in question,
        always return -2 from this method to fall through to normal database authentication.
        Note that if authentication succeeds, murmur will create a record of the user in it's
        database, reserving the username and id so it cannot be used for normal database
        authentication.
        The data in the certificate (name, email addresses etc), as well as the list of signing
        certificates, should only be trusted if certstrong is true.
        Internally, Murmur treats usernames as case-insensitive. It is recommended
        that authenticators do the same. Murmur checks if a username is in use when
        a user connects. If the connecting user is registered, the other username is
        kicked. If the connecting user is not registered, the connecting user is not
        allowed to join the server.
        Arguments:
            name: Username to authenticate.
            pw: Password to authenticate with.
            certificates: List of der encoded certificates the user connected with.
            certhash: Hash of user certificate, as used by murmur internally when matching.
            certstrong: True if certificate was valid and signed by a trusted CA.
            context: The request context for the invocation.
        Returns a tuple containing the following:
        _retval: UserID of authenticated user,
                    -1 for authentication failures,
                    -2 for unknown user (fallthrough),
                    -3 for authentication failures where the data could (temporarily)
                       not be verified.
        newname: Set this to change the username from the supplied one.
        groups: List of groups on the root channel that
                the user will be added to for the duration of the connection.
        """
        if name == "SuperUser":
            debug("Forced fall through for SuperUser")
            return self._authenticate_fallthrough(name)
        try:
            user = self._load_user(where={"username": name})
        except Exception as e:
            error(f"Fetching mumble user failed: {e}")
            return self._authenticate_fallthrough(name)
        if not user:
            return self._authenticate_templink(name, pw)
        uid = user.get("user_id")
        if not uid:
            error(f"User without ID: {user}")
            return self._authenticate_fallthrough(name)
        mumble_user_id = self._to_mumble_baseuser_id(int(uid))
        display_name = user.get("display_name", f"Noname user {uid}")
        upwhash = user.get("pwhash")
        uhashfn = user.get("hashfn")
        groups = user.get("groups", "").split(",")
        if allianceauth_check_hash(pw, upwhash, uhashfn):
            info(f"User authenticated: {display_name} ({mumble_user_id}). Group: {groups}.")
            return self._authenticate_success(name, mumble_user_id, display_name, groups)
        info(
            f"Failed authentication attempt for user: {name} ({uid})",
        )
        return self._authenticate_failure(name)

    def _authenticate_templink(self, name, pw) -> Tuple[int, Optional[str], Optional[List[str]]]:
        """
        Authenticate a user via Templink

        :param name: The user name
        :param pw: The password
        """
        try:
            user = self._load_temp_user(where={"username": name})
        except Exception as e:
            error(f"Fetching Templink user failed: {e}")
            return self._authenticate_fallthrough(name)
        if not user:
            info(f"User {name} not found in Templink")
            return self._authenticate_fallthrough(name)
        password = user.get("password")
        if not password:
            info(f"Templink user {name} has no password")
            return self._authenticate_failure(name)
        if pw != password:
            info(f"Templink user {name} failed password")
            return self._authenticate_failure(name)
        uid = user.get("id")
        if not uid:
            error(f"Templink user without ID: {user}")
            return self._authenticate_fallthrough(name)
        expire = user.get("expires", 0)
        unix_now = time.time()
        if unix_now >= expire:
            info(f"Templink user {name} expired: {expire}")
            return self._authenticate_failure(name)
        display_name = user.get("name", f"Templink User {uid}")
        groups = ["Guest"]
        mumble_user_id = self._to_mumble_tempuser_id(int(uid))
        info(f"User Templink authorized: {display_name} ({mumble_user_id}). Groups: {groups}")
        return self._authenticate_success(name, mumble_user_id, display_name, groups)

    def getInfo(self, id, current=None) -> Tuple[bool, Any]:
        """
        Fetch information about a user. This is used to retrieve information
        like email address, keyhash etc. If you want murmur to take care of this
        information itself, simply return false to fall through.
        Arguments:
            id: User id.
            context: The request context for the invocation.
        Returns a tuple containing the following:
            _retval: true if information is present, false to fall through.
            info: Information about user. This needs to include at least "name".
        """
        if not self._is_baseuser(id):
            debug(f"User {id} not base user, fall through")
            return (GetInfoState.FALLTHROUGH.value, None)
        name = self._id_to_name(id)
        if not name:
            warning(f"User with id {id} not found, fall through")
            return (GetInfoState.FALLTHROUGH.value, None)
        info = {UserInfo.UserName: name}
        debug(f"Present info for user {id}: {info}")
        return (GetInfoState.SUCCESS.value, info)

    def nameToId(self, name, current=None) -> int:
        """
        Map a name to a user id. For example, used in ACLs

        Arguments:
            name: Username to map.
            context: The request context for the invocation.
        Returns: User id or -2 for unknown name.
        """
        debug(f"Starting name to id mapping for name {name}")
        if name.lower() == "superuser":
            debug("Name to id mapping for 'superuser' -> forced fall through")
            return NameToIdState.UNKNOWN.value
        try:
            # Internally, Murmur treats usernames as case-insensitive,
            # so the LIKE operator is used
            user = self._load_user(where={"username": name}, where_operator="LIKE")
        except Exception as e:
            error(f"Failed to fetch the user by username '{name}' from the database: {e}")
            return NameToIdState.UNKNOWN.value
        if not user:
            debug(f"Name {name} to id mapping not found, fall through")
            return NameToIdState.UNKNOWN.value
        uid = user.get("user_id")
        if not uid:
            error(f"User without ID: {user}")
            return NameToIdState.UNKNOWN.value
        mumble_user_id = self._to_mumble_baseuser_id(int(uid))
        debug(f"Mapped name {name} to id {mumble_user_id}")
        return int(mumble_user_id)

    def idToName(self, id, current=None) -> str:
        """
        Map a user id to a username. For example, it is also used in ACLs

        Arguments:
            id: User id to map.
            context: The request context for the invocation.
        Returns: Name of user or empty string for unknown id.
        """
        debug(f"Starting id to name mapping for id {id}")
        if not self._is_baseuser(id):
            debug(f"Id {id} not in baseuser range, fall through")
            return IdToNameState.UNKNOWN.value
        name = self._id_to_name(id)
        if not name:
            warning(f"User id {id} not found, fall through")
            return IdToNameState.UNKNOWN.value
        if name == "superuser":
            warning(f"The 'superuser' ({id}) not allowed, fall through")
            return IdToNameState.UNKNOWN.value
        debug(f"Mapped id {id} to name {name}")
        return name

    def _id_to_name(self, id) -> Optional[str]:
        """
        Makes the Mumble ID to name mapping

        Arguments:
            id: The ID
        Returns: The name if found, otherwise None
        """
        id = self._to_baseuser_id(int(id))
        try:
            user = self._load_user(where={"user_id": id})
        except Exception as e:
            error(f"Failed to fetch the user by ID {id} from the database: {e}")
            return None
        if not user:
            error(f"Failed to find user by ID {id}")
            return None
        # Internally, Murmur treats usernames as case-insensitive,
        # so the lower() function is used
        return str(user.get("username", f"Noname user {id}")).lower()

    def idToTexture(self, id, current=None) -> bytes:
        """
        Map a user to a custom Texture.
        Arguments:
            id: User id to map.
            context: The request context for the invocation.
        Returns: User texture or an empty texture for unknown users or users without textures.
        """

        if not self.cfg["avatar"]["enabled"]:
            info(
                "Avatars are disabled, fall through",
            )
            return IdToTextureState.UNKNOWN.value
        if id >= self.tempuser_offset:
            character_id = self._id_to_character_id_templink(id)
        elif id >= self.base_offset:
            character_id = self._id_to_character_id(id)
        else:
            warning(f"User ID {id} is not in the valid range")
            return IdToTextureState.UNKNOWN.value
        if not character_id:
            warning(f"Character ID for user ID {id} is empty")
            return IdToTextureState.UNKNOWN.value
        return self._avatar_get(character_id)

    def _id_to_character_id(self, id) -> Optional[int]:
        """
        Gets the character ID for the user ID

        :param id: The user ID
        :return: The character ID or None
        """
        user_id = self._to_baseuser_id(int(id))
        try:
            profile = self._load_item_from_database(
                table="authentication_userprofile",
                columns=["main_character_id"],
                where={"user_id": user_id},
            )
        except Exception as e:
            error(f"Failed to fetch the user profile by ID {user_id}: {e}")
            return None
        if not profile:
            error(f"Failed to find user profile by ID {user_id}")
            return None
        main_character_id = profile.get("main_character_id")
        if not main_character_id:
            error(f"User {user_id} not associated with any character")
            return None
        try:
            character = self._load_item_from_database(
                table="eveonline_evecharacter",
                columns=["character_id"],
                where={"id": main_character_id},
            )
        except Exception as e:
            error(f"Failed to fetch the character by user ID {user_id}: {e}")
            return None
        if not character:
            error(f"Failed to find character by user ID {user_id}")
            return None
        character_id = character.get("character_id")
        if not character_id:
            error(f"Character without ID: {character}")
            return None
        return int(character_id)

    def _id_to_character_id_templink(self, id) -> Optional[int]:
        """
        Gets the character ID for the user ID

        :param id: The user ID
        :return: The character ID or None
        """
        user_id = self._to_tempuser_id(int(id))
        try:
            user = self._load_temp_user(where={"id": user_id})
        except Exception as e:
            error(f"Failed to fetch the Templink character by ID {user_id}: {e}")
            return None
        if not user:
            error(f"Failed to find Templink character by ID {user_id}")
            return None
        character_id = user.get("character_id")
        if not character_id:
            error(f"Templink character without ID: {user}")
            return None
        return int(character_id)

    def registerUser(self, name, current=None) -> int:
        """
        Register a new user.
        Arguments:
            info: Information about user to register.
            context: The request context for the invocation.
        Returns: User id of new user, -1 for registration failure, or -2 to fall through.
        """
        self._method_debug("registerUser", name)
        return RegisterUserState.FALLTHROUGH.value

    def unregisterUser(self, id, current=None) -> int:
        """
        Unregister a user.
        Arguments:
            id: Userid to unregister.
            context: The request context for the invocation.
        Returns: 1 for successful unregistration,
                 0 for unsuccessful unregistration,
                 -1 to fall through.
        """
        self._method_debug("unregisterUser", id)
        return UnregisterUserState.FALLTHROUGH.value

    def getRegisteredUsers(self, filter, current=None) -> Dict[int, str]:
        """
        Get a list of registered users matching filter.
        Arguments:
            filter: Substring usernames must contain. If empty, return all registered users.
            context: The request context for the invocation.
        Returns: List of matching registered users.
        """

        if not filter:
            filter = "%"

        try:
            res = self.db.search(
                table="mumble_mumbleuser",
                columns=["user_id", "username"],
                where={"username": filter},
            )
        except Exception as e:
            error(f"Failed to fetch users by substring {filter} from the database: {e}")
            return {}

        if not res:
            debug(f"Empty user list for filter {filter}")
            return {}
        debug(f"Found {len(res)} users for filter {filter}")
        return dict(
            [
                (
                    self._to_mumble_baseuser_id(user.get("user_id")),
                    str(user.get("username")).lower(),
                )
                for user in res
            ]
        )

    def setInfo(self, id, info, current=None) -> int:
        """
        Set additional information for user registration.
        Arguments:
            id: Userid of registered user.
            info: Information to set about user. This should be merged with existing information.
            context: The request context for the invocation.
        Returns: 1 for successful update, 0 for unsuccessful update, -1 to fall through.
        """
        self._method_debug("setInfo", id)
        return SetInfoState.FALLTHROUGH.value

    def setTexture(self, id, texture, current=None) -> int:
        """
        Set texture (now called avatar) of user registration.
        Arguments:
            id: registrationId of registered user.
            tex: New texture.
            context: The request context for the invocation.
        Returns: 1 for successful update, 0 for unsuccessful update, -1 to fall through.
        """
        self._method_debug("setTexture", id)
        return SetTextureState.FALLTHROUGH.value

    def _method_debug(self, fn: str, val: Any) -> None:
        """
        Debugging helper to print the method call.
        """
        debug(f"Process {fn}({val}) -> fall through")



import logging
import os
import sys
from logging import debug, error, info, warning
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml

ENV_PREFIX = "MA__"

DEFAULT_CONFIG_FILE = "authenticator.yml"

DEFAULT_CONFIG = {
    "database": {
        "name": "alliance_auth",
        "user": "allianceserver",
        "password": "password",
        "prefix": "",
        "host": "127.0.0.1",
        "port": 3306,
        "connection_pool_initial_size": 1,
        "connection_pool_max_size": 10,
        "cache_dir": "database_cache",
        "cache_ttl": 5,
        "cache_max_size": 10737418240,  # 10GB
        "cache_max_age": 7776000,  # 90 days
    },
    "user": {
        "id_offset": 1000000000,
        "reject_on_error": True,
    },
    "avatar": {
        "enabled": False,
        "url": "https://images.evetech.net/characters/{character_id}/portrait?size=256",
        "cache_dir": "avatar_cache",
        "cache_max_size": 10737418240,  # 10GB
        "cache_ttl": 2592000,  # 30 days
        "cache_max_age": 7776000,  # 90 days
    },
    "ice_server": {
        "host": "127.0.0.1",
        "port": 6502,
        "secret": "",
    },
    "ice_healthcheck": {
        "enabled": True,
        "check_interval": 1,  # I don't recommend changing
    },
    "ice_client": {
        "bind_address": "127.0.0.1",
        "port": 13004,
    },
    "ice_properties": [],
    "mumble": {
        "servers": [1],
    },
    "glacier": {  # Unsupported!
        "enabled": False,
        "user": "allianceserver",
        "password": "password",
        "host": "localhost",
        "port": 4063,
    },
    "log": {
        "level": "debug",
        "file": None,
        "format": "%(asctime)s %(levelname)s %(module)s %(message)s",
    },
    "idlehandler": {
        "enabled": False,
        "idle_timeout": 3600,
        "check_interval": 600,
        "afk_channel": 0,
        "allowlist": None,
        "denylist": None,
    },
    "prometheus": {
        "enabled": False,
        "port": 8000,
    },
}


class Config:
    def __init__(
        self,
        config_file: Union[str, Path] = DEFAULT_CONFIG_FILE,
        default_config: Dict[str, Any] = DEFAULT_CONFIG,
        env_prefix: str = ENV_PREFIX,
    ) -> None:
        """
        Initialize the class with a default configuration
        and optionally a user configuration file.

        :param config_file: Path to the YAML file with user configuration
        :param default_config: Dictionary with default configuration
        """
        self.env_prefix = env_prefix
        self.default_config = default_config
        self.config = default_config.copy()
        self._logging_config()
        self.config_file = Path(config_file) if isinstance(config_file, str) else config_file
        self._load()
        self._apply_env_vars()
        self._logging_config()

    def _load(self) -> None:
        """
        Load user configuration from a YAML file and merge it with the default configuration.
        """
        if not self.config_file.exists():
            warning(f"Config file '{self.config_file}' not found. Using default values.")
            return

        try:
            with self.config_file.open("r") as file:
                user_config = yaml.safe_load(file)
                self._merge(self.config, user_config or {})
                debug(f"Loaded configuration from {self.config_file}: {self.config}")
        except yaml.YAMLError as e:
            error(f"Error loading YAML from '{self.config_file}': {e}")
            sys.exit(1)
        except Exception as e:
            error(f"Unexpected error loading config '{self.config_file}': {e}")
            sys.exit(1)

    def _merge(self, defaults: Dict[str, Any], custom: Dict[str, Any]) -> None:
        """
        Recursively merge user settings with default settings.

        :param defaults: Dictionary with default settings
        :param custom: Dictionary with user settings
        """
        for key, value in custom.items():
            if isinstance(value, dict) and key in defaults:
                self._merge(defaults[key], value)
            else:
                defaults[key] = value

    def _apply_env_vars(self) -> None:
        """
        Override configuration values using environment variables.
        Environment variables should use the format MA__SECTION__SUBSECTION__KEY.
        List items should be specified as
        MA__SECTION__SUBSECTION__0, MA__SECTION__SUBSECTION__1, etc.
        """
        for env_var, env_value in os.environ.items():
            if not env_var.startswith(self.env_prefix):
                continue
            prefix_size = len(self.env_prefix)
            keys = env_var[prefix_size:].split("__")
            processed_keys: List[Union[str, int]] = []
            for key in keys:
                if key.isdigit():
                    processed_keys.append(int(key))
                else:
                    processed_keys.append(key.lower())
            self._set_nested_value(self.config, processed_keys, env_value)
        debug(f"Loaded configuration from environment: {self.config}")

    def _set_nested_value(
        self, dictionary: Union[Dict[str, Any], list], keys: list, value: Any
    ) -> None:
        """
        Sets a value in an existing multidimensional dictionary using a list of keys.
        Assumes that the dictionary structure already exists according to DEFAULT_CONFIG.

        :param dictionary: The dictionary or list to set the value to.
        :param keys: A list of keys representing the path in the dictionary or list.
        :param value: The value to set on the specified path.
        """
        value = self._convert_value(value)
        for key in keys[:-1]:
            if isinstance(dictionary, dict):
                if key not in dictionary:
                    if isinstance(keys[keys.index(key) + 1], int):
                        dictionary[key] = []
                    else:
                        dictionary[key] = {}
                dictionary = dictionary[key]
            elif isinstance(dictionary, list) and isinstance(key, int):
                while len(dictionary) <= key:
                    next_key = keys[keys.index(key) + 1]
                    if isinstance(next_key, int):
                        dictionary.append([])
                    else:
                        dictionary.append({})
                dictionary = dictionary[key]
            else:
                return
        if isinstance(dictionary, dict):
            dictionary[keys[-1]] = value
        elif isinstance(dictionary, list) and isinstance(keys[-1], int):
            while len(dictionary) <= keys[-1]:
                dictionary.append(None)
            dictionary[keys[-1]] = value

    @staticmethod
    def _convert_value(value: str) -> Optional[Union[int, float, bool, str]]:
        """
        Convert a string value to an appropriate type (int, float, bool, or str).

        :param value: The string value to convert
        :return: The converted value in an appropriate type
        """
        if value.lower() in ("true", "false"):
            return value.lower() == "true"
        if value.lower() == "none":
            return None
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return value

    def export(self, export_file: Union[str, Path] = DEFAULT_CONFIG_FILE) -> None:
        """
        Export the current configuration to a YAML file.

        :param export_file: Path to the file where configuration should be saved
        """
        export_file = Path(export_file) if isinstance(export_file, str) else export_file
        try:
            with export_file.open("w") as file:
                yaml.dump(self.config, file, default_flow_style=False)
                info(f"Configuration exported to '{export_file}'")
        except Exception as e:
            error(f"Failed to export configuration: {e}")

    def _logging_config(self) -> None:
        """
        Get settings for the logging configuration.

        :return: Dictionary with logging settings
        """
        level_mapping = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "warn": logging.WARNING,
            "error": logging.ERROR,
            "err": logging.ERROR,
            "critical": logging.CRITICAL,
            "crit": logging.CRITICAL,
        }
        handler: Union[logging.StreamHandler, logging.FileHandler]
        if self.config["log"]["file"]:
            handler = logging.FileHandler(self.config["log"]["file"])
        else:
            handler = logging.StreamHandler()
        logger = logging.getLogger()
        handler.setFormatter(logging.Formatter(self.config["log"]["format"]))
        level = level_mapping.get(self.config["log"]["level"].lower(), logging.INFO)
        logger.setLevel(level)
        for old_handler in logger.handlers[:]:
            logger.removeHandler(old_handler)
        logger.addHandler(handler)

    def get_database_config(self) -> dict:
        """
        Get settings for the database connection.

        :return: Dictionary with database connection settings
        """
        config = {
            "user": self.config["database"]["user"],
            "password": self.config["database"]["password"],
            "database": self.config["database"]["name"],
        }
        for key in [
            "cache_dir",
            "cache_max_age",
            "cache_max_size",
            "cache_ttl",
            "host",
            "port",
            "prefix",
            "connection_pool_max_size",
            "connection_pool_initial_size",
        ]:
            if self.config["database"][key]:
                config[key] = self.config["database"][key]
        return config



class IDConversionMixin:
    def __init__(self, base_offset: int = 1000000000):
        self.base_offset = base_offset
        self.tempuser_offset = base_offset * 2

    def _to_mumble_baseuser_id(self, user_id: int) -> int:
        """
        Convert user id from database to mumble user id
        """
        return user_id + self.base_offset

    def _to_baseuser_id(self, mubmeble_user_id: int) -> int:
        """
        Convert user id from mumble user id to database user id
        """
        return mubmeble_user_id - self.base_offset

    def _to_mumble_tempuser_id(self, user_id: int) -> int:
        """
        Convert Templink user id from database to mumble user id
        """
        return user_id + self.tempuser_offset

    def _to_tempuser_id(self, mubmeble_user_id: int) -> int:
        """
        Convert user id from mumble user id to Templink database user id
        """
        return mubmeble_user_id - self.tempuser_offset

    def _is_tempuser(self, mubmeble_user_id: int) -> bool:
        """
        Check if the given mubmeble_user_id is for a temporary user.

        Returns True if the ID belongs to a temporary user, False otherwise.
        """
        return mubmeble_user_id >= self.tempuser_offset

    def _is_baseuser(self, mubmeble_user_id: int) -> bool:
        """
        Check if the given mubmeble_user_id is for a base user.

        Returns True if the ID belongs to a regular user, False otherwise.
        """
        return self.base_offset <= mubmeble_user_id < self.tempuser_offset



import time
from logging import debug, error, warning
from typing import Any

import diskcache


class BaseCacheMixin:
    """
    Base class for caching with diskcache.
    """

    def __init__(
        self, cache_dir: str, cache_ttl: int, cache_max_size: int, cache_max_age: int
    ) -> None:
        self.cache_ttl = cache_ttl
        self.cache_expire = cache_max_age
        try:
            self.cache = diskcache.Cache(directory=cache_dir, size_limit=cache_max_size)
            debug(
                "The cache is initialized with parameters:"
                f" dir={cache_dir}, ttl={cache_ttl}, max_size={cache_max_size}"
            )
        except Exception as e:
            error(f"Failed to initialize diskcache at {cache_dir}: {e}")
            self.cache = None

    def _cache_get(self, key: str) -> Any:
        if self.cache is None:
            return (None, None)
        try:
            return self.cache.get(key, default=(None, None))
        except Exception as e:
            warning(f"Failed to retrieve data from cache: {e}")
            return (None, None)

    def _cache_set(self, key: str, value: Any) -> None:
        if self.cache is None:
            return
        try:
            self.cache.set(key, (value, time.time()), expire=self.cache_expire)
        except Exception as e:
            warning(f"Failed to store data in cache: {e}")

    def _cache_is_fresh(self, cache_time: int = 0) -> bool:
        return time.time() - cache_time < self.cache_ttl



import inspect
from logging import debug, error, info

import Ice
from eveo7_mumbleserver_ice import InvalidSecretException
from eveo7_mumbleserver_ice import MetaCallback as MumbleMetaCallback

from .decorators import check_secret, fortify_ice


class MetaCallback(MumbleMetaCallback):
    def __init__(self, app, cfg):
        super().__init__()
        self.app = app
        self.cfg = cfg
        self._decorate_methods()

    def _decorate_methods(self) -> None:
        """
        Applies `check_secret` and `fortify_ice` decorators to all public methods of the class.
        """
        secret = self.cfg["ice_server"]["secret"]
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            # debug(f"Decorating method {name} for meta callback")
            if not name.startswith("_"):
                decorated_method = check_secret(secret=secret)(fortify_ice()(method))
                setattr(self, name, decorated_method)

    def _should_handle_server(self, server_id: int) -> bool:
        """
        Checks if the server with the given server_id should be handled
        based on the configuration.

        :param server_id: ID of the virtual server.
        :return: True if the server should be handled, False otherwise.
        """
        return not self.cfg["mumble"]["servers"] or server_id in self.cfg["mumble"]["servers"]

    def started(self, server, current=None):
        """
        Called when a virtual server starts. Attaches an authenticator if necessary.

        :param server: The virtual server instance.
        :param current: Current Ice context (default: None).
        """
        server_id = server.id()
        if self._should_handle_server(server_id):
            info(f"Setting authenticator for virtual server {server_id}")
            try:
                server.setAuthenticator(self.app.auth)
            # Apparently this server was restarted without us noticing
            except (
                InvalidSecretException,
                Ice.UnknownUserException,
            ) as e:
                error(f"Invalid ice secret: {e}")
                return
        else:
            debug(f"Virtual server {server_id} got started")

    def stopped(self, server, current=None):
        """
        Called when a virtual server stops.

        :param server: The virtual server instance.
        :param current: Current Ice context (default: None).
        """
        if self.app.connected:
            # Only try to output the server id if we think we are still connected to prevent
            # flooding of our thread pool
            try:
                server_id = server.id()
                if self._should_handle_server(server_id):
                    info(f"Authenticated virtual server {server_id} got stopped")
                else:
                    debug(f"Virtual server {server_id} got stopped")
                return
            except Ice.ConnectionRefusedException:
                self.app.connected = False

        debug("Server shutdown stopped a virtual server")



from enum import Enum


class AuthenticationState(Enum):
    """
    The authentication status is used
    to indicate whether a user is authenticated or not.

    Attrtibutes:
        FAILURE (int): Authentication failure
        FALLTHROUGH (int): Unknown user (fallthrough)
        TEMPORARILY (int): Authentication failure where
                           the data could (temporarily) not be verified
    """

    FAILURE = -1
    FALLTHROUGH = -2
    TEMPORARILY = -3


class GetInfoState(Enum):
    """
    The get info status is used to indicate
    if the get info request is successful or not.

    Attrtibutes:
        FAILURE (bool): if information is not present, fall through
        SUCCESS (bool): if information is present
    """

    FALLTHROUGH = False
    SUCCESS = True


class NameToIdState(Enum):
    """
    The name to id status is used to indicate
    if the name to id request is successful or not.

    Attrtibutes:
        UNKNOWN (int): if user is unknown
    """

    UNKNOWN = -2


class IdToNameState(Enum):
    """
    The id to name status is used to indicate
    if the id to name request is successful or not.

    Attrtibutes:
        UNKNOWN (str): if id is unknown
    """

    UNKNOWN = ""


class IdToTextureState(Enum):
    """
    The id to texture status is used to indicate
    if the id to texture request is successful or not.

    Attrtibutes:
        UNKNOWN (byte): if empty texture for unknown users
                        or users without textures.
    """

    UNKNOWN = b""


class RegisterUserState(Enum):
    """
    The register user status is used to indicate
    if the register user request is successful or not.

    Attrtibutes:
        FAILURE (int): if registration failed
        FALLTHROUGH (int): if registration fallthrough
    """

    FAILURE = -1
    FALLTHROUGH = -2


class UnregisterUserState(Enum):
    """
    The unregister user status is used to indicate
    if the unregister user request is successful or not.

    Attrtibutes:
        SUCCESS (int): if unregistration succeeded
        FAILURE (int): if unregistration failed
        FALLTHROUGH (int): if unregistration fallthrough
    """

    SUCCESS = 1
    FAILURE = 0
    FALLTHROUGH = -1


class SetInfoState(Enum):
    """
    The set info status is used to indicate
    if the set info request is successful or not.

    Attrtibutes:
        SUCCESS (int): if set info succeeded
        FAILURE (int): if set info failed
        FALLTHROUGH (int): if set info fallthrough
    """

    SUCCESS = 1
    FAILURE = 0
    FALLTHROUGH = -1


class SetTextureState(Enum):
    """
    The set texture status is used to indicate
    if the set texture request is successful or not.

    Attrtibutes:
        SUCCESS (int): if set texture succeeded
        FAILURE (int): if set texture failed
        FALLTHROUGH (int): if set texture fallthrough
    """

    SUCCESS = 1
    FAILURE = 0
    FALLTHROUGH = -1



from logging import getLogger

from Ice import Logger


class AppLogger(Logger):
    """
    Logger implementation to pipe Ice log messages into
    our own log
    """

    def __init__(self):
        super().__init__()
        self._log = getLogger(__name__)

    def _print(self, message):
        self._log.info(message)

    def trace(self, category, message):
        self._log.debug("Trace %s: %s", category, message)

    def warning(self, message):
        self._log.warning(message)

    def error(self, message):
        self._log.error(message)



from logging import debug, error, warning
from typing import Optional
from urllib.parse import urlparse
from urllib.request import urlopen

from prometheus_client import Counter, Summary

from .mixin_basecache import BaseCacheMixin

AVATAR_CACHE_MISS = Counter("avatar_cache_miss", "Number of cache misses for avatar")
AVATAR_CACHE_HIT = Counter("avatar_cache_hit", "Number of cache hits for avatar")
AVATAR_CACHE_STALE = Counter(
    "avatar_cache_stale", "Number of times stale cache was used for avatar"
)
AVATAR_DOWNLOAD_TIME = Summary("avatar_download_time", "Time spent downloading avatar")


class AvatarCacheMixin(BaseCacheMixin):
    """
    Mixin for handling avatar caching and downloading with diskcache.
    """

    def __init__(
        self,
        url_template: str,
        cache_dir: str = "avatar",
        cache_ttl: int = 2592000,
        cache_max_size: int = 10737418240,
        cache_max_age: int = 7776000,
    ):
        super().__init__(cache_dir, cache_ttl, cache_max_size, cache_max_age)
        self.url_template = url_template

    def _avatar_get(self, character_id: int) -> bytes:
        """
        Retrieves avatar from cache if valid, otherwise downloads and caches it.

        Args:
            character_id (int): Unique character identifier.

        Returns:
            bytes: Avatar image data.
        """
        avatar_data: Optional[bytes] = None
        avatar_url = self.url_template.format(character_id=character_id)
        avatar_data, cache_time = self._cache_get(avatar_url)
        if avatar_data is not None and self._cache_is_fresh(cache_time):
            debug(f"Cache hit for avatar with key '{avatar_url}'")
            AVATAR_CACHE_HIT.inc()
            return avatar_data
        try:
            debug(f"Downloading avatar from URL {avatar_url}")
            avatar_data = self._avatar_download(avatar_url)
            self._cache_set(avatar_url, avatar_data)
            AVATAR_CACHE_MISS.inc()
            return avatar_data
        except Exception as e:
            if avatar_data is None:
                error(f"Failed to download avatar from URL {avatar_url}: {e}")
                return b""
            else:
                warning(f"Returning expired cached avatar for key '{avatar_url}'")
                AVATAR_CACHE_STALE.inc()
                return avatar_data

    @AVATAR_DOWNLOAD_TIME.time()
    def _avatar_download(self, avatar_url: str) -> bytes:
        """
        Downloads the avatar from the specified URL.

        Args:
            avatar_url (str): URL to download the avatar from.

        Returns:
            bytes: Downloaded avatar data.
        """
        parsed_url = urlparse(avatar_url)
        if parsed_url.scheme not in ["http", "https"]:
            raise ValueError(f"Invalid avatar URL scheme: {avatar_url}")
        with urlopen(avatar_url) as response:
            content_type = response.headers.get("Content-Type")
            if not content_type or not content_type.startswith("image/"):
                raise ValueError(f"Invalid content type for avatar: {content_type}")
            avatar_data: bytes = response.read()
            return avatar_data



import datetime
import inspect
from logging import debug, error, info
from typing import Any, Dict

from eveo7_mumbleserver_ice import ServerCallback as MumbleServerCallback

from .db import ConnectionPoolDB
from .decorators import check_secret
from .mixin_idconversion import IDConversionMixin


class ServerCallback(MumbleServerCallback, IDConversionMixin):
    def __init__(self, server, cfg: Dict[str, Any], db: ConnectionPoolDB):
        super().__init__()
        IDConversionMixin.__init__(self, base_offset=cfg["user"]["id_offset"])
        self.server = server
        self.db = db
        self.cfg = cfg
        self._decorate_methods()

    def _decorate_methods(self) -> None:
        secret = self.cfg["ice_server"]["secret"]
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not name.startswith("_"):
                decorated_method = check_secret(secret=secret)(method)
                setattr(self, name, decorated_method)

    def _user_update(self, user, updates: Dict[str, str]) -> None:
        if not self._is_baseuser(user.userid):
            debug(f"Avoid update of user {user.name} with ID {user.userid}")
            return
        try:
            self.db.update(
                table="mumble_mumbleuser",
                updates=updates,
                where={"user_id": self._to_baseuser_id(user.userid)},
            )
        except Exception as e:
            error(
                f"Failed to update user {user.name} in the database: {e}"
                " Maybe you need to update and migrate the database?"
            )

    def userConnected(self, user, current=None):
        self._user_update(
            user,
            updates={
                "release": user.release,
                "version": user.version,
                "last_connect": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            },
        )
        info(f"User {user.name} connected to the server")

    def userDisconnected(self, user, current=None):
        self._user_update(
            user,
            updates={"last_disconnect": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")},
        )
        info(f"User {user.name} disconnected from the server")

    def userStateChanged(self, user, current=None):
        pass

    def channelCreated(self, channel, current=None):
        pass

    def channelRemoved(self, channel, current=None):
        pass

    def channelStateChanged(self, channel, current=None):
        pass

    def userTextMessage(self, user, message, current=None):
        if message.text == "!kicktemps":
            if self.server.hasPermission(user.session, 0, 0x10000):
                self.server.sendMessage(user.session, "Kicking all templink clients!")

                users = self.server.getUsers()
                for _, auser in users.items():
                    if auser.userid > self.tempuser_offset:
                        # print(auser)
                        self.server.kickUser(auser.session, "Kicking all temp users! :-)")

                self.server.sendMessage(user.session, "All templink clients kicked!")

            else:
                self.server.sendMessage(user.session, "You do not have kick permissions!")



from contextlib import contextmanager
from logging import debug, error, warning
from queue import Empty, Full, Queue
from typing import Any, Dict, List, Optional, Tuple

import pymysql
from prometheus_client import Counter, Gauge, Summary
from pymysql.connections import Connection

from .mixin_basecache import BaseCacheMixin

# Metrics for the connection
DB_CONNECTION_SUCCESS = Counter(
    "db_connection_success", "Number of successful connections to the database"
)
DB_CONNECTION_FAILURE = Counter(
    "db_connection_failure", "Number of failed connections to the database"
)
# Metrics for the connection pool
DB_POOL_SIZE = Gauge("db_pool_size", "Current size of the connection pool")
DB_POOL_MAX_SIZE = Gauge("db_pool_max_size", "Maximum size of the connection pool")
DB_POOL_CONNECTION_REUSE = Counter(
    "db_pool_connection_reuse",
    "Number of successful reuses of existing connections from the pool",
)
DB_POOL_CONNECTION_NEW = Counter("db_pool_connection_new", "Number of new connections created")
DB_POOL_CONNECTION_ERROR = Counter(
    "db_pool_connection_error",
    "Number of connection errors when reusing an existing connection",
)
# Metrics for the query
DB_QUERY_SUCCESS = Counter("db_query_success", "Number of successful queries")
DB_QUERY_FAILURE = Counter("db_query_failure", "Number of failed queries")
DB_QUERY_LATENCY = Summary("db_query_latency", "Latency of queries")
# Metrics for the cache
DB_CACHE_SELECT_MISS = Counter(
    "db_cache_select_miss", "Number of cache misses for SELECT queries"
)
DB_CACHE_SELECT_HIT = Counter("db_cache_select_hit", "Number of cache hits for SELECT queries")
DB_CACHE_SELECT_STALE = Counter(
    "db_cache_select_stale", "Number of times stale cache was used for SELECT queries"
)


class ConnectionPoolDBException(Exception):
    pass


class ConnectionPoolDB(BaseCacheMixin):
    """
    Small abstraction to handle database connections for multiple threads.
    """

    def __init__(
        self,
        user: str,
        password: str,
        database: str,
        host: str = "127.0.0.1",
        port: int = 3306,
        prefix: str = "",
        connection_pool_max_size: int = 10,
        connection_pool_initial_size: int = 2,
        cache_dir: str = "database",
        cache_ttl: int = 10,
        cache_max_size: int = 10737418240,
        cache_max_age: int = 7776000,
    ) -> None:
        super().__init__(cache_dir, cache_ttl, cache_max_size, cache_max_age)
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.prefix = prefix
        # Fixed-size connection pool
        DB_POOL_MAX_SIZE.set(connection_pool_max_size)
        self.pool: Queue[Connection] = Queue(maxsize=connection_pool_max_size)
        for _ in range(connection_pool_initial_size):
            try:
                self.pool.put(self._create_new_connection())
            except Exception as e:
                error(f"Could not create connection: {e}")
        DB_POOL_SIZE.set(self.pool.qsize())

    def _create_new_connection(self) -> Connection:
        """
        Creates a new connection to the database.
        """
        try:
            con = pymysql.connect(
                host=self.host,
                port=self.port,
                user=self.user,
                password=self.password,
                database=self.database,
                charset="utf8",
                cursorclass=pymysql.cursors.DictCursor,
                autocommit=True,
            )
            DB_CONNECTION_SUCCESS.inc()
            return con
        except pymysql.MySQLError as e:
            DB_CONNECTION_FAILURE.inc()
            error(f"Could not connect to database: {e}")
            raise ConnectionPoolDBException(str(e))

    def _get_connection(self) -> Connection:
        """
        Gets a connection from the pool or creates a new one if the pool is empty.
        """
        DB_POOL_SIZE.set(self.pool.qsize())
        try:
            con = self.pool.get_nowait()
            # Check if the connection is still active
            try:
                con.ping(reconnect=True)
                DB_POOL_CONNECTION_REUSE.inc()
                debug("Re-using existing connection from pool")
                return con
            except pymysql.MySQLError:
                DB_POOL_CONNECTION_NEW.inc()
                warning("Connection is no longer active, creating a new one")
                con = self._create_new_connection()
            self._return_connection(con)
            return con
        except Empty:
            DB_POOL_CONNECTION_ERROR.inc()
            warning("No available connections in pool, creating a new connection")
            return self._create_new_connection()

    def _return_connection(self, con: Connection) -> None:
        """
        Returns the connection to the pool if it is still active.
        """
        try:
            if con.open:
                self.pool.put_nowait(con)
            else:
                debug("Connection is closed, not returning to pool")
        except Full:
            con.close()
            warning("Connection pool is full, closing connection")
        DB_POOL_SIZE.set(self.pool.qsize())

    @contextmanager
    def _cursor(self) -> Any:
        """
        Context manager for creating and closing cursors.
        """
        connection = self._get_connection()
        cursor = connection.cursor()
        try:
            yield cursor
        except Exception as e:
            error(f"Error occurred during cursor operation: {e}")
            raise
        finally:
            cursor.close()
            self._return_connection(connection)

    @contextmanager
    @DB_QUERY_LATENCY.time()
    def execute(
        self, query: str, params: Optional[Tuple[Any]] = None, max_retries: int = 3
    ) -> Any:
        """
        Execute a database query with retry support.

        :param query: SQL query string.
        :param params: Optional parameters to use with the query.
        :param max_retries: Maximum number of retries in case of failure.
        :return: Result of the executed query.
        """
        retry_count = 0
        while retry_count < max_retries:
            try:
                with self._cursor() as cursor:
                    debug(f"Executing query: {query} with params: {params}")
                    cursor.execute(query, params)
                    DB_QUERY_SUCCESS.inc()
                    yield cursor
                    return
            except pymysql.MySQLError as e:
                DB_QUERY_FAILURE.inc()
                warning(f"Query execution failed: {e}, retrying {retry_count + 1}/{max_retries}")
                retry_count += 1
            if retry_count >= max_retries:
                raise ConnectionPoolDBException(
                    f"Query failed after {max_retries} attempts: {query}"
                )

    def select(
        self,
        table: str,
        columns: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        where_operator: str = "=",
    ) -> Any:
        """
        Execute a SELECT query and return the result.

        :param table: Name of the table.
        :param columns: List of columns to select (defaults to all columns).
        :param where: WHERE conditions as a dictionary.
        :param where_operator: Operator used for WHERE conditions.
        :return: List of tuples with the query results.
        """
        columns_clause = ", ".join([f"`{col}`" for col in columns]) if columns else "*"
        where_clause, params = self._build_where_clause(where, where_operator)
        query = f"SELECT {columns_clause} FROM `{self.table_name(table)}` {where_clause}"
        cache_key = f"{query}:{params}"
        cache_data, cache_time = self._cache_get(cache_key)
        if cache_data is not None and self._cache_is_fresh(cache_time):
            debug(f"Returning data from cache for query {query} with params {params}")
            DB_CACHE_SELECT_HIT.inc()
            return cache_data
        try:
            with self.execute(query, tuple(params)) as cursor:
                result = cursor.fetchall()
                self._cache_set(cache_key, result)
                DB_CACHE_SELECT_MISS.inc()
                return result
        except Exception as e:
            if cache_data is not None:
                DB_CACHE_SELECT_STALE.inc()
                warning(f"Returning stale data from cache due to query failure: {e}")
                return cache_data
            else:
                raise

    def update(
        self,
        table: str,
        updates: Dict[str, Any],
        where: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """
        Execute an UPDATE query and return the number of affected rows.

        :param table: Name of the table.
        :param updates: Dictionary of columns and their new values.
        :param where: WHERE conditions as a dictionary.
        :return: Number of affected rows.
        """
        set_clause = ", ".join([f"`{col}` = %s" for col in updates.keys()])
        params = list(updates.values())
        where_clause, where_params = self._build_where_clause(where)
        params.extend(where_params)

        query = f"UPDATE `{self.table_name(table)}` SET {set_clause} {where_clause}"
        with self.execute(query, tuple(params)) as cursor:
            count = cursor.rowcount
            debug(f"Updated {count} rows in table {table}")
            return count

    def search(
        self,
        table: str,
        columns: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
    ) -> Any:
        return self.select(table=table, columns=columns, where=where, where_operator="LIKE")

    def _build_where_clause(
        self, where: Optional[Dict[str, Any]], operator: str = "="
    ) -> Tuple[str, List[Any]]:
        """
        Build a WHERE clause for a query.

        :param where: Dictionary of WHERE conditions.
        :return: Tuple of WHERE clause and parameter list.
        """
        if not where:
            return "", []

        where_clause = "WHERE " + " AND ".join([f"`{col}` {operator} %s" for col in where.keys()])
        return where_clause, list(where.values())

    def disconnect(self) -> None:
        """
        Closes all connections in the pool.
        """
        while not self.pool.empty():
            connection = self.pool.get()
            connection.close()
            debug("Closing database connection")

    def table_name(self, name: str) -> str:
        """
        Return the full table name, including any prefixes.
        """
        return f"{self.prefix}{name}"



from logging import debug, info, warning
from threading import Timer
from typing import Any, List, Optional

from eveo7_mumbleserver_ice import InvalidChannelException
from prometheus_client import Counter, Summary

AFK_USERS_MOVED_FAILURE = Counter(
    "afk_users_moved_failure", "Number of failed moves of users to AFK channel"
)
AFK_USERS_MOVED_SUCCESS = Counter(
    "afk_users_moved_success", "Number of successful moves of users to AFK channel"
)
AFK_USERS_PROCESS_LATENCY = Summary("afk_users_process_latency", "Time spent processing users")


class IdleHandler:
    """
    Class to handle idle users in the server
    and move them to a specified channel if they are AFK.
    """

    def __init__(
        self,
        server: Any,
        enabled: bool = False,
        idle_timeout: int = 3600,
        check_interval: int = 600,
        afk_channel: int = 1,
        allowlist: List[int] = [],
        denylist: List[int] = [],
    ) -> None:
        """
        Initializes the IdleHandler.

        :param server: Reference to the server object to manage users.
        :param enabled: Whether the handler is enabled.
        :param idle_timeout: Idle time in seconds after which user is considered AFK.
        :param check_interval: Interval in seconds to check for idle users.
        :param afk_channel: The channel ID to move idle users to.
        :param allowlist: List of channel IDs where users are allowed to stay even if idle.
        :param denylist: List of channel IDs where users should not be moved from.
        """
        self.server = server
        self.enabled = enabled
        self.idle_timeout = idle_timeout
        self.check_interval = check_interval
        self.afk_channel = afk_channel
        self.allowlist = allowlist
        self.denylist = denylist
        self._timer: Optional[Timer] = None

    def start(self) -> None:
        """
        Starts the idle handler
        """
        if self.enabled:
            info(f"IdleHandler started for server {self.server.id()}")
            self._process_users()
            self._timer = Timer(self.check_interval, self.start)
            self._timer.start()
        else:
            debug("IdleHandler is disabled")

    def stop(self) -> None:
        """
        Stops the idle handler
        """
        if self._timer:
            self._timer.cancel()

    @AFK_USERS_PROCESS_LATENCY.time()
    def _process_users(self) -> None:
        """
        Fetches and processes all users to check their idle status.
        """
        try:
            users = self.server.getUsers().values()
            debug(f"Fetched All Users: {users}")
            for user in users:
                self._process_user(user)
        except Exception as e:
            warning(f"Failed to fetch or process users: {e}")

    def _process_user(self, user: Any) -> None:
        """
        Processes a user if they are idle and should be moved.

        :param user: The user object to process.
        """
        if isinstance(user, int):
            debug(f"Skipping User {user}. This happens occasionally")
            return
        if user.idlesecs <= self.idle_timeout:
            return
        debug(f"User {user.name} is AFK, for {user.idlesecs}/{self.idle_timeout}")
        try:
            state = self.server.getState(user.session)
            if state and self._should_be_moved(state.channel):
                state.channel = self.afk_channel
                state.selfMute = True
                state.selfDeaf = True
                self.server.setState(state)
                info(f"Moving user {user.name} to the AFK channel {self.afk_channel}.")
                AFK_USERS_MOVED_SUCCESS.inc()
        except InvalidChannelException:
            warning(f"Channel {self.afk_channel} not found.")
            AFK_USERS_MOVED_FAILURE.inc()
        except Exception as e:
            warning(f"Failed to process user {user.name}: {e}")
            AFK_USERS_MOVED_FAILURE.inc()

    def _should_be_moved(self, channel: int) -> bool:
        """
        Checks if the user should be moved to the AFK channel.

        :param channel: The current channel of the user.
        :return: True if the user should be moved, otherwise False.
        """
        # Skip if the user is already in the AFK channel
        if channel == self.afk_channel:
            return False
        # If denylist is used, allowlist is ignored
        if self.denylist is not None:
            if channel in self.denylist:
                return False
            else:
                return True
        if self.allowlist is not None:
            if channel in self.allowlist:
                return True
            else:
                return False
        return True



from logging import debug, error, info, warning
from threading import Timer
from typing import Any, Dict, List, Optional

import Ice
from eveo7_mumbleserver_ice import (
    InvalidSecretException,
    MetaCallbackPrx,
    MetaPrx,
    ServerCallbackPrx,
    ServerUpdatingAuthenticatorPrx,
)
from prometheus_client import Enum

from .authenticator import Authenticator
from .db import ConnectionPoolDB
from .idlehandler import IdleHandler
from .metacallback import MetaCallback
from .servercallback import ServerCallback

ICE_HEALTHCHECK = Enum(
    "ice_healthcheck",
    "Healthcheck for ICE server",
    states=["healthy", "unhealthy", "disabled", "unknown"],
)


class App(Ice.Application):
    def __init__(self, cfg: Dict[str, Any], db: ConnectionPoolDB):
        super().__init__()
        self.db = db
        self.cfg = cfg
        self.meta = None
        self.auth = None
        self.metacb = None
        self.adapter: Ice.ObjectAdapter
        self.connected = False
        self.healthcheck: Optional[Timer] = None
        self._idlehandlers: List[IdleHandler] = []

    def run(self, args):
        self.shutdownOnInterrupt()
        if not self._initialize_ice_connection():
            return 1
        # Start the healthcheck timer if configured
        if self.cfg["ice_healthcheck"]["enabled"]:
            ICE_HEALTHCHECK.state("unknown")
            self._start_healthcheck()
        else:
            ICE_HEALTHCHECK.state("disabled")
        # Wait for the application to shut down
        self.communicator().waitForShutdown()
        # Stop the healthcheck timer if it's running
        if self.healthcheck:
            self.healthcheck.cancel()
        for idlehandler in self._idlehandlers:
            idlehandler.stop()
        if self.interrupted():
            warning("Caught interrupt signal, shutting down application.")
        # Disconnect from the database before exiting
        self.db.disconnect()
        return 0

    def _initialize_ice_connection(self):
        """
        Establishes the Ice connection and sets up the authenticator for servers.
        """
        ice = self.communicator()
        # Set Ice secret if configured
        secret = self.cfg["ice_server"]["secret"]
        glacier_enabled = self.cfg["glacier"]["enabled"]
        if secret:
            debug("Using shared Ice secret.")
            ice.getImplicitContext().put("secret", secret)
        elif not glacier_enabled:
            warning("Consider using an Ice secret to improve security.")
        if glacier_enabled:
            error("Glacier support is not implemented yet.")
            return False
        # Connect to the Ice server
        host = self.cfg["ice_server"]["host"]
        port = self.cfg["ice_server"]["port"]
        info(f"Connecting to Ice server at {host}:{port}.")
        try:
            # Create a proxy and connect to the Meta object
            base = ice.stringToProxy(f"Meta:tcp -h {host} -p {port}")
            self.meta = MetaPrx.uncheckedCast(base)
            # Create an adapter for callbacks
            host = self.cfg["ice_client"]["bind_address"]
            port = self.cfg["ice_client"]["port"]
            self.adapter = ice.createObjectAdapterWithEndpoints(
                "Callback.Client", f"tcp -h {host} -p {port}"
            )
            self.adapter.activate()
            # Set up MetaCallback and Authenticator proxies
            self.metacb = MetaCallbackPrx.uncheckedCast(
                self.adapter.addWithUUID(MetaCallback(self, self.cfg))
            )
            self.auth = ServerUpdatingAuthenticatorPrx.uncheckedCast(
                self.adapter.addWithUUID(Authenticator(self.cfg, self.db))
            )
            return self._attach_callbacks()
        except Exception as e:
            self._handle_ice_exception(e)
            return False

    def _attach_callbacks(self):
        """
        Attaches all necessary callbacks to Meta and the authenticator.
        """
        try:
            debug("Attaching Meta callback.")
            # Attach Meta callback
            self.meta.addCallback(self.metacb)
            # Attach callbacks for each virtual server
            for server in self.meta.getBootedServers():
                if (
                    not self.cfg["mumble"]["servers"]
                    or server.id() in self.cfg["mumble"]["servers"]
                ):
                    self._attach_callbacks_for_server(server)
        except Exception as e:
            self._handle_ice_exception(e)
            return False
        self.connected = True
        return True

    def _attach_callbacks_for_server(self, server) -> None:
        """
        Attaches callbacks to a virtual server.

        :param server: The virtual server to attach callbacks to.
        """
        server_id = server.id()
        info(f"Setting authenticator for virtual server {server_id}.")
        # Create and attach server callback
        servercb = ServerCallbackPrx.uncheckedCast(
            self.adapter.addWithUUID(ServerCallback(server, self.cfg, self.db))
        )
        server.setAuthenticator(self.auth)
        server.addCallback(servercb)
        # Start IdleHandler for the server
        idlehandler = IdleHandler(server, **self.cfg["idlehandler"])
        idlehandler.start()
        self._idlehandlers.append(idlehandler)

    def _reattach(self):
        """
        Attempts to reconnect to the Ice server upon connection closure or exception.
        """
        info("Reconnecting to Ice server...")
        if not self._attach_callbacks():
            error(
                "Reconnection attempt failed. Will retry in next %s seconds.",
                self.cfg["ice_healthcheck"]["check_interval"],
            )

    def _start_healthcheck(self) -> None:
        """
        Starts the healthcheck timer to monitor the connection state periodically.
        """
        self.healthcheck = Timer(
            self.cfg["ice_healthcheck"]["check_interval"], self._start_healthcheck
        )
        self.healthcheck.start()
        self._check_connection()

    def _check_connection(self) -> None:
        """
        Checks the connection state and reconnects if needed.

        I haven't found any other adequate way to check
        the connection other than pinging every second.
        """
        if not self.meta:
            self._initialize_ice_connection()
            return
        try:
            self.meta.ice_ping()
            if not self.connected:
                error("Connection is inactive. Attempting to reconnect...")
                ICE_HEALTHCHECK.state("unhealthy")
                self._reattach()
            else:
                debug("Connection is active.")
                ICE_HEALTHCHECK.state("healthy")
        except Ice.ConnectionLostException:
            error("Connection lost. Attempting to reconnect...")
            ICE_HEALTHCHECK.state("unhealthy")
            self._reattach()
        except Ice.Exception as e:
            error(f"Failed to ping the Ice server: {e}. Trying to reconnect...")
            ICE_HEALTHCHECK.state("unhealthy")
            self._reattach()

    def _handle_ice_exception(self, e: Exception) -> None:
        """
        Handles Ice-related exceptions and logs appropriate messages.
        """
        if isinstance(e, Ice.ConnectionRefusedException):
            error("Connection refused by server.")
        elif isinstance(e, (InvalidSecretException, Ice.UnknownUserException)):
            error("Invalid Ice secret or unknown user.")
        else:
            error(f"Unexpected Ice exception: {e}")
            raise e
        self.connected = False



import argparse
import sys
import threading
from logging import debug, info
from typing import Any, Dict

import Ice
from prometheus_client import start_http_server

from .app import App
from .config import DEFAULT_CONFIG_FILE, Config
from .db import ConnectionPoolDB
from .logger import AppLogger


def initialize_ice_properties(cfg: Dict[str, Any]) -> Ice.InitializationData:
    initdata = Ice.InitializationData()
    initdata.properties = Ice.createProperties([], initdata.properties)
    for line in cfg.get("ice_properties", []):
        prop, val = line.split("=", 2)
        if prop and val:
            debug(f"Ice property {prop}={val}")
            initdata.properties.setProperty(prop.strip(), val.strip())
    initdata.properties.setProperty("Ice.ImplicitContext", "Shared")
    initdata.logger = AppLogger()
    return initdata


def start_metrics_server(port: int) -> None:
    start_http_server(port)


def main() -> None:
    parser = argparse.ArgumentParser(description="AllianceAuth Mumble Authenticator")
    parser.add_argument(
        "config_file",
        type=str,
        help="Path to config file",
        default=DEFAULT_CONFIG_FILE,
        nargs="?",
    )
    parser.add_argument(
        "--default-config",
        type=str,
        help="Export default config to YAML file. If this flag is used, all other arguments will be ignored.",
    )
    args = parser.parse_args()
    if args.default_config:
        config = Config()
        config.export(args.default_config)
        sys.exit()

    config = Config(args.config_file)
    if config.config["prometheus"]["enabled"]:
        debug(f"Starting metrics server on port {config.config['prometheus']['port']}")
        metrics_thread = threading.Thread(
            target=start_metrics_server, args=(config.config["prometheus"]["port"],)
        )
        metrics_thread.start()
    info("Starting AllianceAuth Mumble authenticator")
    app = App(config.config, ConnectionPoolDB(**config.get_database_config()))
    state = app.main(sys.argv[:1], initData=initialize_ice_properties(config.config))
    info(f"Shutdown complete with state {state}")
    sys.exit(state)


if __name__ == "__main__":
    main()



from functools import wraps
from logging import critical
from typing import Any, Callable, Optional, Tuple, Type

import Ice
from eveo7_mumbleserver_ice import InvalidSecretException


def check_secret(secret: Optional[str] = None) -> Callable[[Callable], Callable]:
    """
    Parameterized decorator that checks whether the server transmitted the right secret
    if a secret is supposed to be used.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            current = kwargs.get("current", args[-1] if args else None)

            if secret and (not current or current.ctx.get("secret") != secret):
                critical("Server transmitted invalid secret. Possible injection attempt.")
                raise InvalidSecretException()

            return func(*args, **kwargs)

        return wrapper

    return decorator


def fortify_ice(
    retval: Optional[Any] = None, exceptions: Tuple[Type[Exception], ...] = (Ice.Exception,)
) -> Callable[[Callable], Callable]:
    """
    Decorator that catches exceptions, logs them, and returns a safe retval
    value. This helps prevent the authenticator from getting stuck in
    critical code paths. Only exceptions that are instances of classes
    given in the exceptions list are not caught.

    The default is to catch all non-Ice exceptions.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kws: Any) -> Any:
            try:
                return func(*args, **kws)
            except exceptions:
                raise
            except Exception as e:
                critical(f"Unexpected exception in {func.__name__}: {e}")
                return retval

        return wrapper

    return decorator
